@@ -13,19 +13,24 @@ import logging
 import struct
 import hashlib
 import os
-from typing import Optional
-from io import BytesIO
+import typing
+from typing import Optional, Tuple
 from urllib.parse import urlparse, parse_qs
 
 from pymonetdb.exceptions import OperationalError, DatabaseError, \
     ProgrammingError, NotSupportedError, IntegrityError
 
+if typing.TYPE_CHECKING:
+    from pymonetdb.filetransfer.downloads import Downloader
+    from pymonetdb.filetransfer.uploads import Uploader
+
 logger = logging.getLogger(__name__)
 
 MAX_PACKAGE_LENGTH = (1024 * 8) - 2
 
 MSG_PROMPT = ""
 MSG_MORE = "\1\2\n"
+MSG_FILETRANS = "\1\3\n"
 MSG_INFO = "#"
 MSG_ERROR = "!"
 MSG_Q = "&"
@@ -41,6 +46,8 @@ MSG_TUPLE_NOSLICE = "="
 MSG_REDIRECT = "^"
 MSG_OK = "=OK"
 
+MSG_FILETRANS_B = bytes(MSG_FILETRANS, 'utf-8')
+
 STATE_INIT = 0
 STATE_READY = 1
 
@@ -89,14 +96,17 @@ class Connection(object):
         self.hostname = ""
         self.port = 0
         self.username = ""
-        self.password = ""
         self.database = ""
         self.language = ""
         self.handshake_options = None
         self.connect_timeout = socket.getdefaulttimeout()
+        self.uploader = None
+        self.downloader = None
+        self.stashed_buffer = None
 
-    def connect(self, database, username, password, language, hostname=None,
-                port=None, unix_socket=None, connect_timeout=-1, handshake_options=None):
+    def connect(self, database: str, username: str, password: str, language: str,  # noqa: C901
+                hostname: Optional[str] = None, port: Optional[int] = None, unix_socket=None, connect_timeout=-1,
+                handshake_options=None):
         """ setup connection to MAPI server
 
         unix_socket is used if hostname is not defined.
@@ -104,7 +114,8 @@ class Connection(object):
 
         if ':' in database:
             if not database.startswith('mapi:monetdb:'):
-                raise DatabaseError("colon not allowed in database name, except as part of mapi:monetdb://<hostname>[:<port>]/<database> URI")
+                raise DatabaseError("colon not allowed in database name, except as part of "
+                                    "mapi:monetdb://<hostname>[:<port>]/<database> URI")
             parsed = urlparse(database[5:])
             # parse basic settings
             if parsed.hostname or parsed.port:
@@ -135,10 +146,10 @@ class Connection(object):
                 # Future work: parse other parameters such as reply_size.
 
         if hostname and hostname[:1] == '/' and not unix_socket:
-            unix_socket = '%s/.s.monetdb.%d' % (hostname, port)
+            unix_socket = f'{hostname}/.s.monetdb.{port}'
             hostname = None
-        if not unix_socket and os.path.exists("/tmp/.s.monetdb.%i" % port):
-            unix_socket = "/tmp/.s.monetdb.%i" % port
+        if not unix_socket and os.path.exists(f"/tmp/.s.monetdb.{port}"):
+            unix_socket = f"/tmp/.s.monetdb.{port}"
         elif not unix_socket and not hostname:
             hostname = 'localhost'
 
@@ -150,7 +161,6 @@ class Connection(object):
         self.hostname = hostname
         self.port = port
         self.username = username
-        self.password = password
         self.database = database
         self.language = language
         self.unix_socket = unix_socket
@@ -191,17 +201,17 @@ class Connection(object):
 
         if not (self.language == 'control' and not self.hostname):
             # control doesn't require authentication over socket
-            self._login()
+            self._login(password=password)
 
         self.socket.settimeout(socket.getdefaulttimeout())
         self.state = STATE_READY
 
-    def _login(self, iteration=0):
+    def _login(self, password: str, iteration=0):
         """ Reads challenge from line, generate response and check if
         everything is okay """
 
         challenge = self._getblock()
-        response = self._challenge_response(challenge)
+        response = self._challenge_response(challenge, password)
         self._putblock(response)
         prompt = self._getblock().strip()
 
@@ -224,7 +234,7 @@ class Connection(object):
             if redirect[1] == "merovingian":
                 logger.debug("restarting authentication")
                 if iteration <= 10:
-                    self._login(iteration=iteration + 1)
+                    self._login(iteration=iteration + 1, password=password)
                 else:
                     raise OperationalError("maximal number of redirects "
                                            "reached (10)")
@@ -235,9 +245,10 @@ class Connection(object):
                 self.port = int(self.port)
                 logger.info("redirect to monetdb://%s:%s/%s" %
                             (self.hostname, self.port, self.database))
-                self.socket.close()
+                if self.socket:
+                    self.socket.close()
                 self.connect(hostname=self.hostname, port=self.port,
-                             username=self.username, password=self.password,
+                             username=self.username, password=password,
                              database=self.database, language=self.language)
 
             else:
@@ -252,15 +263,32 @@ class Connection(object):
         self.state = STATE_INIT
         self.socket.close()
 
-    def cmd(self, operation):
+    def _sabotage(self):
+        """ Kill the connection in a way that the server is sure to recognize as an error"""
+        sock = self.socket
+        self.socket = None
+        self.state = STATE_INIT
+        if not sock:
+            return
+        bad_header = struct.pack('<H', 2 * 8193 + 0)  # larger than allowed, and not the final message
+        bad_body = b"ERROR\x80ERROR"  # invalid utf-8, and too small
+        try:
+            sock.send(bad_header + bad_body)
+            # and then we hang up
+            sock.close()
+        except Exception:
+            # don't care
+            pass
+
+    def cmd(self, operation):  # noqa: C901
         """ put a mapi command on the line"""
         logger.debug("executing command %s" % operation)
 
         if self.state != STATE_READY:
-            raise (ProgrammingError, "Not connected")
+            raise ProgrammingError("Not connected")
 
         self._putblock(operation)
-        response = self._getblock()
+        response = self._getblock_and_transfer_files()
         if not len(response):
             return ""
         elif response.startswith(MSG_OK):
@@ -297,7 +325,7 @@ class Connection(object):
         else:
             raise ProgrammingError("unknown state: %s" % response)
 
-    def _challenge_response(self, challenge):
+    def _challenge_response(self, challenge: str, password: str):  # noqa: C901
         """ generate a response to a mapi login challenge """
 
         challenges = challenge.split(':')
@@ -306,7 +334,6 @@ class Connection(object):
         challenges.pop()
 
         salt, identity, protocol, hashes, endian = challenges[:5]
-        password = self.password
 
         if protocol == '9':
             algo = challenges[5]
@@ -319,15 +346,15 @@ class Connection(object):
         else:
             raise NotSupportedError("We only speak protocol v9")
 
-        for h in hashes.split(","):
+        for i in hashes.split(","):
             try:
-                s = hashlib.new(h)
+                s = hashlib.new(i)
             except ValueError:
                 pass
             else:
                 s.update(password.encode())
                 s.update(salt.encode())
-                pwhash = "{" + h + "}" + s.hexdigest()
+                pwhash = "{" + i + "}" + s.hexdigest()
                 break
         else:
             raise NotSupportedError("Unsupported hash algorithms required"
@@ -336,6 +363,7 @@ class Connection(object):
         response = ":".join(["BIG", self.username, pwhash, self.language, self.database]) + ":"
 
         if len(challenges) >= 7:
+            response += "FILETRANS:"
             options_level = 0
             for part in challenges[6].split(","):
                 if part.startswith("sql="):
@@ -352,63 +380,107 @@ class Connection(object):
 
         return response
 
-    def _getblock(self):
-        """ read one mapi encoded block """
-        if self.language == 'control' and not self.hostname:
-            return self._getblock_socket()  # control doesn't do block splitting when using a socket
-        else:
-            return self._getblock_inet()
+    def _getblock_and_transfer_files(self):
+        """ read one mapi encoded block and take care of any file transfers the server requests"""
+        buffer = self._get_buffer()
+        offset = 0
 
-    def _getblock_inet(self):
-        result = BytesIO()
-        last = 0
-        while not last:
-            flag = self._getbytes(2)
-            unpacked = struct.unpack('<H', flag)[0]  # little endian short
-            length = unpacked >> 1
-            last = unpacked & 1
-            result.write(self._getbytes(length))
-        return result.getvalue().decode()
+        # import this here to solve circular import
+        from pymonetdb.filetransfer import handle_file_transfer
 
-    def _getblock_socket(self):
-        buffer = BytesIO()
         while True:
-            x = self.socket.recv(1)
-            if len(x):
-                buffer.write(x)
+            old_offset = offset
+            offset = self._getblock_raw(buffer, old_offset)
+            i = buffer.rfind(b'\n', old_offset, offset - 1)
+            if i >= old_offset + 2 and buffer[i - 2: i + 1] == MSG_FILETRANS_B:
+                # File transfer request. Chop the cmd off the buffer by lowering the offset
+                cmd = str(buffer[i + 1: offset - 1], 'utf-8')
+                offset = i - 2
+                handle_file_transfer(self, cmd)
+                continue
             else:
                 break
-        return buffer.getvalue().strip().decode()
+        self._stash_buffer(buffer)
+        return str(memoryview(buffer)[:offset], 'utf-8')
 
-    def _getbytes(self, bytes_):
-        """Read an amount of bytes from the socket"""
-        result = BytesIO()
-        count = bytes_
-        while count > 0:
-            recv = self.socket.recv(count)
-            if len(recv) == 0:
+    def _getblock(self) -> str:
+        """ read one mapi encoded block """
+        buf = self._get_buffer()
+        end = self._getblock_raw(buf, 0)
+        ret = str(memoryview(buf)[:end], 'utf-8')
+        self._stash_buffer(buf)
+        return ret
+
+    def _getblock_raw(self, buffer: bytearray, offset: int) -> int:
+        """
+        Read one mapi block into 'buffer' starting at 'offset', enlarging the buffer
+        as necessary and returning offset plus the number of bytes read.
+        """
+        last = False
+        while not last:
+            offset, last = self._get_minor_block(buffer, offset)
+        return offset
+
+    def _get_minor_block(self, buffer: bytearray, offset: int) -> Tuple[int, bool]:
+        self._getbytes(buffer, offset, 2)
+        unpacked = buffer[offset] + 256 * buffer[offset + 1]
+        length = unpacked >> 1
+        last = unpacked & 1
+        if length:
+            offset = self._getbytes(buffer, offset, length)
+        return (offset, bool(last))
+
+    def _getbytes(self, buffer: bytearray, offset: int, count: int) -> int:
+        """
+        Read 'count' bytes from the socket into 'buffer' starting at 'offset'.
+        Enlarge buffer if necessary.
+        Return offset + count if all goes well.
+        """
+        assert self.socket
+        end = count + offset
+        if len(buffer) < end:
+            # enlarge
+            nblocks = 1 + (end - len(buffer)) // 8192
+            buffer += bytes(nblocks * 8192)
+        while offset < end:
+            view = memoryview(buffer)[offset:end]
+            n = self.socket.recv_into(view)
+            if n == 0:
                 raise BrokenPipeError("Server closed connection")
-            count -= len(recv)
-            result.write(recv)
-        return result.getvalue()
+            offset += n
+        return end
+
+    def _get_buffer(self) -> bytearray:
+        """Retrieve a previously stashed buffer for reuse, or create a new one"""
+        if self.stashed_buffer:
+            buffer = self.stashed_buffer
+            self.stashed_buffer = None
+        else:
+            buffer = bytearray(8192)
+        return buffer
+
+    def _stash_buffer(self, buffer):
+        """Stash a used buffer for future reuse"""
+        if self.stashed_buffer is None or len(self.stashed_buffer) < len(buffer):
+            self.stashed_buffer = buffer
 
     def _putblock(self, block):
         """ wrap the line in mapi format and put it into the socket """
-        if self.language == 'control' and not self.hostname:
-            return self.socket.send(block.encode())  # control doesn't do block splitting when using a socket
-        else:
-            self._putblock_inet(block)
+        self._putblock_inet_raw(block.encode(), True)
 
-    def _putblock_inet(self, block):
+    def _putblock_raw(self, block, finish: bool):
+        """ put the data into the socket """
+        self._putblock_inet_raw(block, finish)
+
+    def _putblock_inet_raw(self, block, finish):
         pos = 0
         last = 0
-        block = block.encode()
         while not last:
-            data = block[pos:pos + MAX_PACKAGE_LENGTH]
+            data = memoryview(block)[pos:pos + MAX_PACKAGE_LENGTH]
             length = len(data)
             if length < MAX_PACKAGE_LENGTH:
                 last = 1
-            flag = struct.pack('<H', (length << 1) + last)
+            flag = struct.pack('<H', (length << 1) + (last if finish else 0))
             self.socket.send(flag)
             self.socket.send(data)
             pos += length
@@ -428,6 +500,14 @@ class Connection(object):
 
         self.cmd("Xreply_size %s" % size)
 
+    def set_uploader(self, uploader: "Uploader"):
+        """Register the given Uploader, or None to deregister"""
+        self.uploader = uploader
+
+    def set_downloader(self, downloader: "Downloader"):
+        """Register the given Downloader, or None to deregister"""
+        self.downloader = downloader
+
 
 # When all supported Python versions support it we can enable @dataclass here.
 class HandshakeOption:
